import ast
import copy
import time as ttime
from datetime import datetime
import pprint
import argparse
import enum
import os
import yaml

import bluesky_queueserver
from .comms import zmq_single_request, validate_zmq_key, generate_zmq_public_key, generate_new_zmq_key_pair

import logging

logger = logging.getLogger(__name__)

qserver_version = bluesky_queueserver.__version__


class CommandParameterError(Exception):
    ...


class QServerExitCodes(enum.Enum):
    SUCCESS = 0
    PARAMETER_ERROR = 1
    REQUEST_FAILED = 2
    COMMUNICATION_ERROR = 3
    EXCEPTION_OCCURRED = 4


default_user = "qserver-cli"
default_user_group = "admin"

# The following text is displayed as part of help information (-h or --help option)
cli_examples = """
Examples of CLI commands
------------------------
qserver -h       # Display help
qserver monitor  # Start 'qserver' in monitoring mode

qserver ping     # Send 'ping' request to RE Manager via ZMQ
qserver status   # Request status of RE Manager

qserver environment open         # Open RE environment
qserver environment close        # Close RE environment
qserver environment destroy      # Destroy RE environment (kill RE worker process)

qserver existing plans           # Request the list of existing plans
qserver existing devices         # Request the list of existing devices
qserver allowed plans            # Request the list of allowed plans
qserver allowed devices          # Request the list of allowed devices
qserver permissions reload       # Reload user permissions and generate lists of allowed plans and devices.
qserver permissions reload lists # Same, but reload lists of existing plans and devices from disk.

qserver permissions set <path-to-file>  # Set user group permissions (from .yaml file)
qserver permissions get                 # Get current user group permissions

qserver queue add plan '<plan-params>'                 # Add plan to the back of the queue
qserver queue add instruction <instruction>            # Add instruction to the back of the queue
qserver queue add plan front '<plan-params>'           # Add plan to the front of the queue
qserver queue add plan back '<plan-params>'            # Add plan to the back of the queue
qserver queue add plan 2 '<plan-params>'               # Insert plan at position 2
qserver queue add instruction 2 <instruction>          # Insert instruction at position 2
qserver queue add plan -1 '<plan-params>'              # Insert plan at position -1
qserver queue add plan before '<uid>' '<plan-params>'  # Insert the plan before the plan with given UID
qserver queue add plan after '<uid>' '<plan-params>'   # Insert the plan after the plan with given UID
NOTE: Position indices are 0-based. Inserting a plan to position 0 pushes it to the front of the queue.
      Negative position indices are counted from the back of the queue. Request for a plan with index -1
      returns the last plan of the queue. Inserting a plan at position -1 makes it previous to last.

qserver queue update plan <uid> '<plan-params>'         #  Update item with <uid> with a plan
qserver queue replace plan <uid> '<plan-params>'        #  Replace item with <uid> with a plan
qserver queue update instruction <uid> '<instruction>'  #  Update item with <uid> with an instruction
qserver queue replace instruction <uid> '<instruction>' #  Replace item with <uid> with an instruction

qserver queue execute plan '<plan-params>'              # Immediately execute the plan
qserver queue execute instruction <instruction>         # Immediately execute an instruction

Example of JSON specification of a plan:
    '{"name": "count", "args": [["det1", "det2"]], "kwargs": {"num": 10, "delay": 1}}'

Supported queue instructions:
    queue-stop  # stops execution of the queue

qserver queue get    # Request the list of items (plans or instructions) in the queue
qserver queue clear  # Clear the queue (remove all plans from the queue)

qserver queue item get           # Request the last item in the queue
qserver queue item get back      # Request the last item in the queue
qserver queue item get front     # Request the first item in the queue
qserver queue item get 2         # Request the item at position 2
qserver queue item get '<uid>'   # Request the item with given Item UID

qserver queue item remove          # Remove the last item from the queue
qserver queue item remove back     # Remove the last item from the queue
qserver queue item remove front    # Remove the first item from the queue
qserver queue item remove 2        # Remove the item at position 2
qserver queue item remove '<uid>'  # Remove the item with the given UID

qserver queue item move 2 5                             # Move item from position 2 to position 5 of the queue
qserver queue item move back front                      # Move item from the back to the front of the queue
qserver queue item move front -2                        # Move item from the front of the queue to position -2
qserver queue item move '<uid-src>' 5                   # Move item with UID <uid-src> to position 5
qserver queue item move 2 before '<uid-dest>'           # Place item at position 2 before an item with <uid-dest>
qserver queue item move 2 after '<uid-dest>'            # Place item at position 2 after an item with <uid-dest>
qserver queue item move '<uid-src>' before '<uid-dest>' # Place item with <uid-src> before item with <uid-dest>

qserver queue start        # Start execution of the queue
qserver queue stop         # Request execition of the queue to stop after current plan
qserver queue stop cancel  # Cancel request to stop execution of the queue

# Queue can operate in LOOP mode, which is disabled by default. To enable or disable the LOOP mode use
qserver queue mode set loop True
qserver queue mode set loop False

# The following requests are forwarded to the Run Engine:
qserver re pause           # Request to PAUSE currently executed plan at the next checkpoint
qserver re pause deferred  # Request to PAUSE currently executed plan at the next checkpoint
qserver re pause immediate # Request to immediately PAUSE currently executed plan
qserver re resume          # RESUME execution of a paused plan
qserver re stop            # STOP execution of a paused plan
qserver re abort           # ABORT execution of a paused plan
qserver re halt            # HALT execution of a paused plan

qserver re runs            # Get the list of active runs (runs generated by the currently running plans)
qserver re runs active     # Get the list of active runs
qserver re runs open       # Get the list of open runs (subset of active runs)
qserver re runs closed     # Get the list of closed runs (subset of active runs)

qserver history get        # Request plan history
qserver history clear      # Clear plan history

qserver function execute <function-params>             # Start execution of a function
qserver function execute <function-params> background  # ... in the background thread

Example of JSON specification of a function ("args" and "kwargs" are optional):
    '{"name": "function_sleep", "args": [20], "kwargs": {}}'

qserver script upload <path-to-file>              # Upload a script to RE Worker environment
qserver script upload <path-to-file> background   # ... in the background
qserver script upload <path-to-file> update-re    # ... allow 'RE' and 'db' to be updated

qserver task result <task-uid>  # Load status or result of a task with the given UID

qserver manager stop           # Safely exit RE Manager application
qserver manager stop safe on   # Safely exit RE Manager application
qserver manager stop safe off  # Force RE Manager application to stop
NOTE: Exit with 'safe on' option will succeed only if RE Manager is in IDLE state (queue is not running).
If called with 'safe off' option, the request will force RE Manager to terminate RE Worker process and
exit even if a plan is running.

qserver manager kill test  # Kills RE Manager by stopping asyncio event loop. Used only for testing.
"""


def extract_source_address(params):
    """
    Extract 'source' item address (index or UID) from the list of parameters. Returns
    the list of remaining parameters. The source address is represented by 1 parameter:
    keywords ``front``, ``back``, or integer number represent a positional address (index),
    any other string is interpreted as UID.

    Parameters
    ----------
    params : list(str)
         List of parameters. The first parameter is interpreted as a source address.

    Returns
    -------
    dict
        Dictionary that contains source address. Elements: ``pos`` - positional address
        (value is int or a string from the set ``front``, ``back``), ``uid`` - uid of
        the item (value is a string representing UID). Empty dictionary if no address
        is found.
    list(str)
        List of the remaining parameters

    Raises
    ------
    IndexError
        Insufficient number of parameters is provided (less than 1)
    """
    n_used = 0

    pos, uid = None, None
    if params[0] in ("front", "back"):
        pos = params[0]
        n_used = 1
    else:
        try:
            pos = int(params[0])
            n_used = 1
        except Exception:
            ...

        if pos is None:
            uid = params[0]
            n_used = 1

    if pos is not None:
        addr_param = {"pos": pos}
    elif uid is not None:
        addr_param = {"uid": uid}
    else:
        addr_param = {}
    return addr_param, params[n_used:]


def extract_destination_address(params):
    """
    Extract 'destination' item address (index or UID) from the list of parameters. Returns
    the list of remaining parameters. The source address is represented by 1 or 2 parameters:
    if 1st parameter is a keywords ``front``, ``back``, or integer number, then the parameter
    represents a positional address (index); if the 1st parameter is a keyword ``before`` or
    ``after``, the 2nd parameter is considered to represent item UID (string). If the 1st
    parameter can not be converted to ``int`` or equal to one of the keywords, it is considered
    that the address is not found.

    Parameters
    ----------
    params : list(str)
         List of parameters. The 1st and optionally the 2nd parameters are interpreted as a destination
         address.

    Returns
    -------
    dict
        Dictionary that contains destination address. Elements: ``pos`` - positional address
        (value is int or a string from the set ``front``, ``back``), ``before_uid`` or ``after_uid``
        - uid of the item preceding or following the destination for the item in the queue (value
        is a string representing UID).  Empty dictionary if no address is found.
    list(str)
        List of the remaining parameters.

    Raises
    ------
    IndexError
        Insufficient number of parameters is provided.
    """
    n_used = 0

    pos, uid, uid_key = None, None, None
    if params[0] in ("front", "back"):
        pos = params[0]
        n_used = 1
    elif params[0] in ("before", "after"):
        uid = params[1]
        # Keys are "before_uid" and "after_uid"
        uid_key = f"{params[0]}_uid"
        n_used = 2
    else:
        try:
            pos = int(params[0])
            n_used = 1
        except Exception:
            ...

    if pos is not None:
        addr_param = {"pos": pos}
    elif uid is not None:
        addr_param = {uid_key: uid}
    else:
        addr_param = {}
    return addr_param, params[n_used:]


def format_list_as_command(params):
    """
    Format list of items as a string. Restores the look of the parameters as they appear in command
    line. Used for printing error messages.

    Parameters
    ----------
    params : list
        List of parameters to print

    Returns
    -------
    str
        Representation of the list as a formatted string
    """
    return " ".join([str(_) for _ in params])


def raise_request_not_supported(params):
    """
    Raises ``CommandParameterError`` exception with ``request is not supported`` message.

    Parameters
    ----------
    params : list
        List of parameters that represent the request. The request will be included in the error
        message

    Raises
    ------
    CommandParameterError
    """
    s = format_list_as_command(params)
    raise CommandParameterError(f"Request '{s}' is not supported")


def check_number_of_parameters(params, n_min, n_max, params_report=None):
    """
    Checks if the number of parameters in ``params`` list is in the range ``[n_min, n_max]``.
    Raises exception if the number of parameters is less than ``n_min`` or more than ``n_max``.

    Parameters
    ----------
    params : list
        The list of parameters. If the number of parameters in the list is outside the
        range ``[n_min, n_max]``, then ``CommandParameterError`` exception is raised.
    n_min, n_max : int
        The range for the number of parameters in ``params`` list.
    params_report : list
        The list of parameters that is included in the error message. If the value is
        ``None``, then parameters are not included in the error message.

    Raises
    ------
    CommandParameterError
    """
    s = format_list_as_command(params_report) if params_report is not None else None
    if len(params) < n_min:
        err_msg = "Some parameters are missing in request"
        if params_report is not None:
            err_msg += f" '{s}'"
        if params_report == params:
            err_msg += f": Minimum number of parameters: {n_min}"
        raise CommandParameterError(err_msg)
    if len(params) > n_max:
        err_msg = "Request"
        if params_report is not None:
            err_msg += f" '{s}'"
        err_msg += " contains extra parameters"
        if params_report == params:
            err_msg += f": Minimum number of parameters: {n_max}"
        raise CommandParameterError(err_msg)


def msg_queue_add_update(params, *, cmd_opt):
    """
    Generate outgoing message for `queue add` command. See ``cli_examples`` for supported formats
    for the command.

    Parameters
    ----------
    params : list
        List of parameters of the command. The first element of the list is expected to be ``add`` keyword.
    cmd_opt : str
        Command option, must match ``param[0]``.

    Returns
    -------
    str
        Name of the method from RE Manager API
    dict
        Dictionary of the method parameters

    Raises
    ------
    CommandParameterError
    """
    # Check if the function was called for the appropriate command
    command = "queue"
    expected_p0 = cmd_opt
    if params[0] != expected_p0:
        raise ValueError(f"Incorrect parameter value '{params[0]}'. Expected value: '{expected_p0}'")

    # Make sure that there is sufficient number of parameters to start processing
    if len(params) < 3:
        raise CommandParameterError(f"Item type and value are not specified: '{command} {params[0]}'")

    if (params[0] in ("update", "replace")) and (len(params) != 4):
        raise CommandParameterError(f"Incorrect number of parameters: '{command} {params[0]}'")

    p_item_type = params[1]
    if p_item_type not in ("plan", "instruction"):
        raise_request_not_supported([command, params[0], params[1]])

    try:
        # Destination address is optional. If no destination index or UID found, then
        #   'addr_param' is {}.
        if params[0] == "add":
            update_uid = None
            addr_param, p_item = extract_destination_address(params[2:])
        elif params[0] in ("update", "replace"):
            update_uid = params[2]  # Next parameter is UID
            addr_param, p_item = {}, params[3:]
        elif params[0] == "execute":
            update_uid = None
            addr_param, p_item = {}, params[2:]
        else:
            raise CommandParameterError(f"Option '{params[0]}' is not supported: '{command} {params[0]}'")

        # There should be exactly 1 parameter left. This parameter should contain a plan
        #   or an instruction.
        check_number_of_parameters(p_item, 1, 1, params)
        if p_item_type == "plan":
            try:
                # Convert quoted string to dictionary.
                plan = ast.literal_eval(p_item[0])
            except Exception:
                raise CommandParameterError(f"Error occurred while parsing the plan '{p_item[0]}'")
            if update_uid:
                plan["item_uid"] = update_uid
            plan["item_type"] = "plan"
            addr_param.update({"item": plan})
        elif p_item_type == "instruction":
            if p_item[0] == "queue-stop":
                instruction = {"name": "queue_stop"}
            else:
                raise CommandParameterError(f"Unsupported instruction type: {p_item[0]}")
            if update_uid:
                instruction["item_uid"] = update_uid
            instruction["item_type"] = "instruction"
            addr_param.update({"item": instruction})
        else:
            # This indicates a bug in the program.
            raise ValueError(f"Unknown item type: {p_item_type}")

    except IndexError:
        raise CommandParameterError(f"The command '{params}' contains insufficient number of parameters")

    option = params[0] if (params[0] != "replace") else "update"
    method = f"{command}_item_{option}"
    prms = addr_param
    prms["user"] = default_user
    prms["user_group"] = default_user_group
    if params[0] in ("update", "replace"):
        prms["replace"] = params[0] == "replace"

    return method, prms


def msg_function_execute(params, *, cmd_opt):
    """
    Generate outgoing message for `function_execute` command. See ``cli_examples`` for supported formats
    for the command.

    Parameters
    ----------
    params : list
        List of parameters of the command. The first elements of the list is expected to be
        ``["function", "execute"]``.
    cmd_opt : str
        Command option, must match ``param[0]``.

    Returns
    -------
    str
        Name of the method from RE Manager API
    dict
        Dictionary of the method parameters

    Raises
    ------
    CommandParameterError
    """
    # Check if the function was called for the appropriate command
    command = "function"
    expected_p0 = cmd_opt
    if params[0] != expected_p0:
        raise ValueError(f"Incorrect parameter value '{params[0]}'. Expected value: '{expected_p0}'")

    call_params = {}

    try:
        if params[0] == "execute":
            if len(params) < 2:
                raise CommandParameterError(f"Incorrect number of parameters: '{command} {params[0]}'")

            item_str = params[1]
            try:
                # Convert quoted string to dictionary.
                item = ast.literal_eval(item_str)
            except Exception:
                raise CommandParameterError(f"Error occurred while parsing the plan '{item_str}'")

            item["item_type"] = "function"

            run_in_background = False
            allowed_params = ("background",)
            for p in params[2:]:
                if p not in allowed_params:
                    raise CommandParameterError(f"Unsupported combination of parameters: '{command} {params}'")
                if p == "background":
                    run_in_background = True

            call_params["item"] = item
            call_params["run_in_background"] = run_in_background
        else:
            raise CommandParameterError(f"Option '{params[0]}' is not supported: '{command} {params[0]}'")

    except IndexError:
        raise CommandParameterError(f"The command '{params}' contains insufficient number of parameters")

    method = f"{command}_{params[0]}"
    prms = call_params
    prms["user"] = default_user
    prms["user_group"] = default_user_group

    return method, prms


def msg_queue_item(params):
    """
    Generate outgoing message for `queue item` command. The supported options are ``get``,
    ``move`` and ``remove``.

    Parameters
    ----------
    params : list
        List of parameters of the command. The first element of the list is expected to be ``get`` keyword.

    Returns
    -------
    str
        Name of the method from RE Manager API
    dict
        Dictionary of the method parameters

    """
    # Check if the function was called for the appropriate command
    command = "queue"
    expected_p0 = "item"
    if params[0] != expected_p0:
        raise ValueError(f"Incorrect parameter value '{params[0]}'. Expected value: '{expected_p0}'")

    # Make sure that there is a sufficient number of parameters to start processing
    if len(params) < 2:
        raise CommandParameterError(f"Item type and options are not specified '{command} {params[0]}'")

    p_item_type = params[1]
    if p_item_type not in ("get", "remove", "move"):
        raise_request_not_supported([command, params[0], params[1]])
    try:
        if p_item_type in ("get", "remove"):
            if len(params) >= 3:
                addr_param_src, p_item = extract_source_address(params[2:])
                # There should be no parameters left
                check_number_of_parameters(p_item, 0, 0, params)
                if not addr_param_src:
                    raise CommandParameterError(
                        f"Source address could not be found: '{format_list_as_command(params)}'"
                    )

                addr_param = addr_param_src
            else:
                # Default: operation with the element at the back of the queue.
                #   We don't need to pass the item address explicitly.
                addr_param = {}

        elif p_item_type == "move":
            addr_param_src, p_item = extract_source_address(params[2:])
            addr_param_dest, p_item = extract_destination_address(p_item)

            # There should be no parameters left
            check_number_of_parameters(p_item, 0, 0, params)
            if not addr_param_src:
                raise CommandParameterError(
                    f"Source index or UID is not found: '{format_list_as_command(params)}'"
                )
            if not addr_param_dest:
                raise CommandParameterError(
                    f"Destination aindex or UID is not found: '{format_list_as_command(params)}'"
                )

            # Change the key from 'pos' to 'pos_dest' ('pos' is used for 'source' position)
            pos_dest = addr_param_dest.pop("pos", None)
            if pos_dest is not None:
                addr_param_dest["pos_dest"] = pos_dest

            addr_param = addr_param_src
            addr_param.update(addr_param_dest)

        else:
            # This indicates a bug in the program.
            raise ValueError(f"Unknown item type: {p_item_type}")

    except IndexError:
        raise CommandParameterError(f"The command '{params}' contain insufficient number of parameters")

    method = f"{command}_{params[0]}_{params[1]}"
    prms = addr_param

    return method, prms


def msg_queue_mode(params):
    """
    Generate outgoing messages for `queue_mode_...` commands. The supported option is ``set``.

    Parameters
    ----------
    params : list
        List of parameters of the command. The first two elements of the list are expected to
        be ``mode`` and ``set`` keywords.

    Returns
    -------
    str
        Name of the method from RE Manager API
    dict
        Dictionary of the method parameters
    """
    # Check if the function was called for the appropriate command
    command = "queue"
    expected_p0 = "mode"
    if params[0] != expected_p0:
        raise ValueError(f"Incorrect parameter value '{params[0]}'. Expected value: '{expected_p0}'")

    # Make sure that there is a sufficient number of parameters to start processing
    if len(params) < 2:
        raise CommandParameterError(f"Item type and options are not specified '{command} {params[0]}'")

    p_item_type = params[1]
    if p_item_type != "set":
        raise_request_not_supported([command, params[0], params[1]])

    try:
        if p_item_type == "set":
            params_mode = params[2:]
            if len(params_mode) % 2:
                raise CommandParameterError(
                    f"The list of queue mode parameters must have even number of elements: {params_mode}"
                )

            queue_mode = {params_mode[i]: params_mode[i + 1] for i in range(0, len(params_mode), 2)}
            for k in queue_mode.keys():
                # Attempt to evaluate key parameters (e.g. "True" should become boolean True)
                #   If a parameter fails to evaluate, it should remain a string.
                try:
                    queue_mode[k] = eval(queue_mode[k], {}, {})
                except Exception:
                    pass
            cmd_prms = {"mode": queue_mode}
        else:
            # This indicates a bug in the program.
            raise ValueError(f"Unknown item type: {p_item_type}")

    except IndexError:
        raise CommandParameterError(f"The command '{params}' contain insufficient number of parameters")

    method = f"{command}_{params[0]}_{params[1]}"
    prms = cmd_prms

    return method, prms


def msg_queue_stop(params):
    """
    Generate outgoing message for `queue stop` command.

    Parameters
    ----------
    params : list
        List of parameters of the command. The first element of the list is expected to be ``stop`` keyword.

    Returns
    -------
    str
        Name of the method from RE Manager API
    dict
        Dictionary of the method parameters

    """
    # Check if the function was called for the appropriate command
    command = "queue"
    expected_p0 = "stop"
    if params[0] != expected_p0:
        raise ValueError(f"Incorrect parameter value '{params[0]}'. Expected value: '{expected_p0}'")

    check_number_of_parameters(params, 1, 2, params)

    method = f"{command}_{params[0]}"
    if len(params) == 2:
        if params[1] == "cancel":
            method = f"{command}_{params[0]}_{params[1]}"
        else:
            cmd = format_list_as_command([command] + params)
            raise CommandParameterError(f"Unknown option '{params[1]}' in the command '{cmd}'")

    prms = {}
    return method, prms


def msg_re_pause(params):
    """
    Generate outgoing message for `re pause` command.

    Parameters
    ----------
    params : list
        List of parameters of the command. The first element of the list is expected to be ``pause`` keyword.

    Returns
    -------
    str
        Name of the method from RE Manager API
    dict
        Dictionary of the method parameters

    """
    # Check if the function was called for the appropriate command
    command = "re"
    expected_p0 = "pause"
    if params[0] != expected_p0:
        raise ValueError(f"Incorrect parameter value '{params[0]}'. Expected value: '{expected_p0}'")

    check_number_of_parameters(params, 1, 2, params)

    method = f"{command}_{params[0]}"
    option = "deferred"
    if len(params) == 2:
        if params[1] in ("deferred", "immediate"):
            option = params[1]
        else:
            cmd = format_list_as_command([command] + params)
            raise CommandParameterError(f"Unknown option '{params[1]}' in the command '{cmd}'")

    prms = {"option": option}

    return method, prms


def create_msg(params):
    """
    Create outgoing message based on command-line arguments.

    Parameters
    ----------
    params : list
        List of parameters (positional arguments) supplied via command line.

    Returns
    -------
    method : str
        The name of the method.
    prms : dict
        The dictionary of method parameters.
    monitoring_mode : bool
        Indicates if monitoring mode is enabled.

    Raises
    ------
    CommandParameterError
        Failure to generate consistent outgoing message based on provided list of parameters.
    """
    if not params:
        raise CommandParameterError("Command is not specified")

    if not isinstance(params, list):
        raise TypeError("Command arguments are not represented as a list")

    monitoring_mode = False

    command = params[0]
    params = params[1:]

    prms = {}

    # ----------- monitor ------------
    if command == "monitor":
        if len(params) != 0:
            raise CommandParameterError(f"Parameters are not allowed for '{command}' request")
        monitoring_mode = True
        method = "ping"

    # ----------- ping ------------
    elif command == "ping":
        if len(params) != 0:
            raise CommandParameterError(f"Parameters are not allowed for '{command}' request")
        method = command

    # ----------- status ------------
    elif command == "status":
        if len(params) != 0:
            raise CommandParameterError(f"Parameters are not allowed for '{command}' request")
        method = command

    # ----------- environment ------------
    elif command == "environment":
        if len(params) != 1:
            raise CommandParameterError(f"Request '{command}' must include only one parameter")
        supported_params = ("open", "close", "destroy")
        if params[0] in supported_params:
            method = f"{command}_{params[0]}"
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- allowed ------------
    elif command == "allowed":
        if len(params) != 1:
            raise CommandParameterError(f"Request '{command}' must include only one parameter")
        supported_params = ("plans", "devices")
        if params[0] in supported_params:
            method = f"{params[0]}_{command}"
            prms["user_group"] = default_user_group
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- existing ------------
    elif command == "existing":
        if len(params) != 1:
            raise CommandParameterError(f"Request '{command}' must include only one parameter")
        supported_params = ("plans", "devices")
        if params[0] in supported_params:
            method = f"{params[0]}_{command}"
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- permissions ------------
    elif command == "permissions":
        if params[0] == "reload":
            if len(params) not in (1, 2):
                raise CommandParameterError(
                    f"Request '{command} {params[0]}' must include only one or two parameters"
                )
            method = f"{command}_{params[0]}"
            if len(params) == 2:
                if params[1] == "lists":
                    prms["reload_plans_devices"] = True
                else:
                    CommandParameterError(f"Request '{command} {params[0]} {params[1]}' is not supported")

        elif params[0] == "get":
            if len(params) != 1:
                raise CommandParameterError(f"Request '{command} {params[0]}' must include only one parameter")
            method = f"{command}_{params[0]}"

        elif params[0] == "set":
            if len(params) != 2:
                raise CommandParameterError(f"Request '{command} {params[0]}' must include only one parameter")
            method = f"{command}_{params[0]}"

            file_name = params[1]
            try:
                with open(file_name, "r") as f:
                    permissions_yaml = f.read()
                permissions_dict = yaml.load(permissions_yaml, Loader=yaml.FullLoader)
            except Exception as ex:
                raise CommandParameterError(
                    f"Request '{command}': failed to read the user group permissions from file '{file_name}': {ex}"
                )
            prms["user_group_permissions"] = permissions_dict

        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- script ------------
    elif command == "script":
        if len(params) < 2:
            raise CommandParameterError(f"Request '{command}' must include at least two parameters")
        if params[0] == "upload":
            # Parameter 1 (required) - file name, parameter 2 (optional) - 'background'
            file_name = params[1]
            try:
                with open(file_name, "r") as f:
                    script = f.read()
            except Exception as ex:
                raise CommandParameterError(
                    f"Request '{command}': failed to read the script from file '{file_name}': {ex}"
                )

            run_in_background = False
            update_re = False
            allowed_values = ("background", "update-re")

            for p in params[2:]:
                if p not in allowed_values:
                    raise CommandParameterError(
                        f"Request '{command}': parameter combination {params} is not supported"
                    )
                if p == "background":
                    run_in_background = True
                elif p == "update-re":
                    update_re = True

            method = f"{command}_{params[0]}"
            prms = {"script": script, "run_in_background": run_in_background, "update_re": update_re}

        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- task ------------
    elif command == "task":
        if len(params) != 2:
            raise CommandParameterError(f"Request '{command}' must include at 3 parameters")
        if params[0] == "result":
            task_uid = str(params[1])
            method = f"{command}_{params[0]}"
            prms = {"task_uid": task_uid}
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- function ------------
    elif command == "function":
        if len(params) < 1:
            raise CommandParameterError(f"Request '{command}' must include at least one parameter")
        if params[0] == "execute":
            method, prms = msg_function_execute(params, cmd_opt=params[0])
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- queue ------------
    elif command == "queue":
        if len(params) < 1:
            raise CommandParameterError(f"Request '{command}' must include at least one parameter")
        supported_params = ("add", "update", "replace", "execute", "get", "clear", "item", "start", "stop", "mode")
        if params[0] in supported_params:
            if params[0] in ("add", "update", "replace", "execute"):
                method, prms = msg_queue_add_update(params, cmd_opt=params[0])

            elif params[0] in ("get", "clear", "start"):
                if len(params) != 1:
                    raise CommandParameterError(f"Request '{command} {params[0]}' must include only one parameter")
                method, prms = f"{command}_{params[0]}", {}

            elif params[0] == "stop":
                method, prms = msg_queue_stop(params)

            elif params[0] == "item":
                method, prms = msg_queue_item(params)

            elif params[0] == "mode":
                method, prms = msg_queue_mode(params)

        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    # ----------- re ------------
    elif command == "re":
        if len(params) < 1:
            raise CommandParameterError(f"Request '{command}' must include at least one parameter")
        supported_params = ("pause", "resume", "stop", "abort", "halt", "runs")
        if params[0] in supported_params:
            if params[0] == "pause":
                method, prms = msg_re_pause(params)
            elif params[0] == "runs":
                if len(params) == 1:
                    method, prms = f"{command}_{params[0]}", {}
                elif (len(params) == 2) and (params[1] in ("active", "open", "closed")):
                    method, prms = f"{command}_{params[0]}", {"option": params[1]}
                else:
                    raise CommandParameterError(
                        f"Unrecognized combination of parameters: {format_list_as_command(params)}"
                    )
            else:
                method, prms = f"{command}_{params[0]}", {}
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    elif command == "history":
        if len(params) < 1:
            raise CommandParameterError(f"Request '{command}' must include at least one parameter")
        supported_params = ("get", "clear")
        if params[0] in supported_params:
            method, prms = f"{command}_{params[0]}", {}
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    elif command == "manager":
        if len(params) < 1:
            raise CommandParameterError(f"Request '{command}' must include at least one parameter")
        if params[0] == "stop":
            method = f"{command}_{params[0]}"
            if len(params) == 1:
                prms = {}
            elif (len(params) == 3) and (params[1] == "safe") and (params[2] in ("on", "off")):
                option = f"{params[1]}_{params[2]}"
                prms = {"option": option}
            else:
                raise CommandParameterError(
                    f"Unsupported number or combination of parameters: {format_list_as_command(params)}"
                )
        elif params[0] == "kill":
            if params == ["kill", "test"]:
                method = f"{command}_{params[0]}"
                prms = {}
            else:
                raise CommandParameterError(
                    f"Unsupported number or combination of parameters: {format_list_as_command(params)}"
                )
        else:
            raise CommandParameterError(f"Request '{command} {params[0]}' is not supported")

    else:
        raise CommandParameterError(f"Unrecognized command '{command}'")

    return method, prms, monitoring_mode


def prepare_qserver_output(msg):
    """
    Prepare received message for printing by 'qserver' CLI tool.

    Parameters
    ----------
    msg: dict
        Full message received from RE Manager.

    Returns
    -------
    str
        Formatted message, which is ready for printing.
    """
    msg = copy.deepcopy(msg)
    # Do not print large dicts in the log: replace values with "..."
    large_dicts = ("plans_allowed", "plans_existing", "devices_allowed", "devices_existing")
    for dict_name in large_dicts:
        if dict_name in msg:
            d = msg[dict_name]
            for k in d.keys():
                d[k] = "{...}"
    return pprint.pformat(msg)


def qserver():

    logging.basicConfig(level=logging.WARNING)
    logging.getLogger("bluesky_queueserver").setLevel("ERROR")

    s_enc = (
        "If RE Manager is configured to use encrypted ZeroMQ communication channel,\n"
        "the encryption must also be enabled before running 'qserver' CLI tool by setting\n"
        "the environment variable QSERVER_ZMQ_PUBLIC_KEY to the value of a valid public key\n"
        "(z85-encoded 40 character string):\n\n"
        "    export QSERVER_ZMQ_PUBLIC_KEY='<public_key>'\n\n"
        "Encryption is disabled by default."
    )

    def formatter(prog):
        # Set maximum width such that printed help mostly fits in the RTD theme code block (documentation).
        return argparse.RawDescriptionHelpFormatter(prog, max_help_position=20, width=90)

    parser = argparse.ArgumentParser(
        description="Command-line tool for communicating with RE Monitor.\n"
        f"bluesky-queueserver version {qserver_version}.\n",
        formatter_class=formatter,
        epilog=f"\n\n{s_enc}\n\n{cli_examples}\n\n",
    )

    parser.add_argument(
        "command",
        metavar="command",
        type=str,
        nargs="+",
        help="a sequence of keywords and parameters that define the command",
    )
    parser.add_argument(
        "--address",
        "-a",
        dest="address",
        action="store",
        default="tcp://localhost:60615",
        help="Address of the server, e.g. 'tcp://127.0.0.1:60615' (default: '%(default)s').",
    )

    args = parser.parse_args()
    print(f"Arguments: {args.command}")

    try:
        # Read public key from the environment variable, then check if the CLI parameter exists
        zmq_public_key = os.environ.get("QSERVER_ZMQ_PUBLIC_KEY", None)
        zmq_public_key = zmq_public_key if zmq_public_key else None  # Case of key==""
        if zmq_public_key is not None:
            try:
                validate_zmq_key(zmq_public_key)
            except Exception as ex:
                raise CommandParameterError(f"ZMQ public key is improperly formatted: {ex}")

        method, params, monitoring_mode = create_msg(args.command)
        if monitoring_mode:
            print("Running QServer monitor. Press Ctrl-C to exit ...")

        while True:
            msg, msg_err = zmq_single_request(
                method, params, zmq_server_address=args.address, server_public_key=zmq_public_key
            )

            now = datetime.now()
            current_time = now.strftime("%H:%M:%S")

            if not msg_err:
                print(f"{current_time} - MESSAGE: \n{prepare_qserver_output(msg)}")
                if isinstance(msg, dict) and ("success" in msg) and (msg["success"] is False):
                    exit_code = QServerExitCodes.REQUEST_FAILED
                else:
                    exit_code = QServerExitCodes.SUCCESS
            else:
                print(f"{current_time} - ERROR: {msg_err}")
                exit_code = QServerExitCodes.COMMUNICATION_ERROR

            if not monitoring_mode:
                break
            ttime.sleep(1)

    except CommandParameterError as ex:
        logger.error("Invalid command or parameters: %s.", str(ex))
        exit_code = QServerExitCodes.PARAMETER_ERROR
    except Exception as ex:
        logger.exception("Exception occurred: %s.", str(ex))
        exit_code = QServerExitCodes.EXCEPTION_OCCURRED
    except KeyboardInterrupt:
        print("\nThe program was manually stopped.")
        exit_code = QServerExitCodes.SUCCESS

    return exit_code.value


def qserver_zmq_keys():

    logging.basicConfig(level=logging.WARNING)
    logging.getLogger("bluesky_queueserver").setLevel("INFO")

    def formatter(prog):
        # Set maximum width such that printed help mostly fits in the RTD theme code block (documentation).
        return argparse.RawDescriptionHelpFormatter(prog, max_help_position=20, width=90)

    parser = argparse.ArgumentParser(
        description="Bluesky-QServer:\nZMQ security: Generate public-private key pair for "
        f"ZeroMQ control communication channel.\nbluesky-queueserver version {qserver_version}.\n\n"
        f"Generate new public-private key pair for secured 0MQ control connection between\n"
        f"RE Manager and client applications. If private key is passed as ``--zmq-private-key``\n"
        f"parameter, then the generated key pair is based on the provided private key.\n",
        formatter_class=formatter,
    )
    parser.add_argument(
        "--zmq-private-key",
        dest="zmq_private_key",
        type=str,
        default=None,
        help="Private key used by RE Manager. If the private key is provided, then the public "
        "key is generated based on the private key. This option allows to create (recover) "
        "public key based on known private key. The passed value should be 40 character "
        "string containing z85 encrypted key.",
    )

    args = parser.parse_args()
    try:
        if args.zmq_private_key is not None:
            private_key = args.zmq_private_key
            validate_zmq_key(private_key)  # Will generate nice error message in case the key is invalid
            public_key = generate_zmq_public_key(private_key)
            msg = "Private key generated based on provided private key."
        else:
            public_key, private_key = generate_new_zmq_key_pair()
            msg = "New public-private key pair."

        print("====================================================================================")
        print(f"     ZMQ SECURITY: {msg}")
        print("====================================================================================")
        print(f" Private key (RE Manager):                 {private_key}")
        print(f" Public key (CLI client or HTTP server):   {public_key}")
        print("====================================================================================\n")

    except Exception as ex:
        print(f"Failed to generate keys: {ex}")
        return QServerExitCodes.EXCEPTION_OCCURRED.value

    return QServerExitCodes.SUCCESS.value



import pytest

from bluesky_queueserver.manager.comms import zmq_single_request

from .common import re_manager_cmd, db_catalog  # noqa: F401

from .common import (
    wait_for_condition,
    condition_environment_created,
    condition_manager_idle,
    condition_environment_closed,
)

# User name and user group name used throughout most of the tests.
_user, _user_group = "Testing Script", "admin"


@pytest.mark.xfail(reason="For some reason the test fails when run on CI, but expected to pass locally")
def test_fixture_db_catalog(db_catalog):  # noqa F811
    """
    Basic test for the fixture `db_catalog`.
    """
    assert db_catalog["catalog_name"] == "qserver_tests"
    # Catalog does not necessarily need to be empty, since it will accumulate results of
    #   all the tests in the current session.
    list(db_catalog["catalog"])

    # Try to instantiate the Data Broker
    from databroker import Broker

    Broker.named(db_catalog["catalog_name"])

    # Try to access the catalog in 'standard' way
    from databroker import catalog

    catalog.force_reload()
    assert list(catalog[db_catalog["catalog_name"]]) == list(db_catalog["catalog"])


def test_fixture_re_manager_cmd_1(re_manager_cmd):  # noqa F811
    """
    Basic test for ``re_manager_cmd``.
    """
    # Create RE Manager with no parameters
    re_manager_cmd()
    # The second call is supposed to close the existing RE Manager and create a new one.
    re_manager_cmd([])


def test_fixture_re_manager_cmd_2(re_manager_cmd, db_catalog):  # noqa F811
    """
    Test for the fixture ``re_manager_cmd``: start RE Manager with command line parameters.
    Subscribe RE to databroker (created by ``db_catalog``, execute the plan and make sure
    that the run is recorded by the databroker by comparing Run UIDs from history and
    start document.
    """
    db_name = db_catalog["catalog_name"]
    re_manager_cmd(["--databroker-config", db_name])

    cat = db_catalog["catalog"]

    plan = {"name": "scan", "args": [["det1", "det2"], "motor", -1, 1, 10], "item_type": "plan"}

    # Plan
    params1 = {"item": plan, "user": _user, "user_group": _user_group}
    resp1, _ = zmq_single_request("queue_item_add", params1)
    assert resp1["success"] is True, f"resp={resp1}"

    resp2, _ = zmq_single_request("status")
    assert resp2["items_in_queue"] == 1
    assert resp2["items_in_history"] == 0

    # Open the environment.
    resp3, _ = zmq_single_request("environment_open")
    assert resp3["success"] is True
    assert wait_for_condition(time=10, condition=condition_environment_created)

    resp4, _ = zmq_single_request("queue_start")
    assert resp4["success"] is True

    assert wait_for_condition(time=5, condition=condition_manager_idle)

    resp5, _ = zmq_single_request("status")
    assert resp5["items_in_queue"] == 0
    assert resp5["items_in_history"] == 1

    resp6, _ = zmq_single_request("history_get")
    history = resp6["items"]
    assert len(history) == 1

    uid = history[-1]["result"]["run_uids"][0]
    start_doc = cat[uid].metadata["start"]
    assert start_doc["uid"] == uid

    # Close the environment.
    resp7, _ = zmq_single_request("environment_close")
    assert resp7["success"] is True, f"resp={resp7}"
    assert wait_for_condition(time=5, condition=condition_environment_closed)



import functools
import inspect
import jsonschema
import copy

_parameter_annotation_schema = {
    "type": "object",
    "properties": {
        "description": {"type": "string"},
        "parameters": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "description": {"type": "string"},
                    "annotation": {"type": "string"},
                    "devices": {
                        "$ref": "#/definitions/custom_types_plans_devices",
                    },
                    "plans": {
                        "$ref": "#/definitions/custom_types_plans_devices",
                    },
                    "enums": {
                        "$ref": "#/definitions/custom_types",
                    },
                    "default": {},  # Expression of any time should be accepted
                    "min": {"type": "number"},
                    "max": {"type": "number"},
                    "step": {"type": "number"},
                    "convert_plan_names": {"type": "boolean"},
                    "convert_device_names": {"type": "boolean"},
                },
            },
        },
    },
    "additionalProperties": False,
    "definitions": {
        "custom_types": {
            "type": "object",
            "additionalProperties": False,
            "patternProperties": {
                "^[_a-zA-Z][_a-zA-Z0-9]*$": {
                    "type": "array",
                    "items": {"type": "string"},
                },
            },
        },
        "custom_types_plans_devices": {
            "type": "object",
            "additionalProperties": False,
            "patternProperties": {
                "^[_a-zA-Z][_a-zA-Z0-9]*$": {
                    "anyOf": [
                        {
                            "type": "array",
                            "items": {"type": "string"},
                        },
                    ],
                },
            },
        },
    },
}


def parameter_annotation_decorator(annotation):
    """
    The decorator allows to attach a custom description to a function or generator function.
    The purpose of the decorator is to extend the standard Python annotation
    capabilities. The annotation elements defined in the decorator always override
    the identical elements defined in the function header (type hints and default values) and
    the docstring (text descriptions of the function and its parameters). All the fields in
    the dictionary passed to the decorator are optional, so only the parameters that are needed
    should be specified. In many cases the complete description of the function may be generated
    from the existing docstring and parameter hints. In those cases using the decorator may be avoided.

    The purpose of the custom parameter annotations is to provide alternative type hints for some
    parameters. Those alternative hints may be necessary for validation of plan parameters outside
    the Run Engine environment. Standard Python types, such as ``float``, ``int`` or ``str``),
    may be unambiguously specified as parameter hints and used for plan validation and typechecking
    (e.g. with ``mypy``). Some plans may accept references to devices or other Bluesky plans
    defined in RE Worker namespace as parameters. Plans and devices are referred by their names
    in parameters of the plans submitted to the queue. Type validation for those parameters
    must include checking the parameter type (e.g. if the parameter is a list) and verifying
    that the values are strings from specified set (e.g. that strings in the list are the devices
    from specified set). This functionality may be achieved by defining custom enum types that
    are supported by the decorator. Custom enum types allow define list of names for ``devices``
    (names of devices from RE Worker namespace), ``plans`` (names of plans from RE Worker namespace)
    and ``enums`` (string literals). Type annotations in the decorator may use custom enum types
    to define complex types (using generic types based on ``typing`` module).

    The decorator allows to define additional parameter annotation items: ``min``, ``max`` and
    ``step``. Those items are applied only to numbers in data structures passed as parameter values.
    If ``min`` or ``max`` values are defined for a parameter, each number found
    in the data structure (such as list, tuple, dictionary, list of dictionaries etc.)
    is compared to the values of ``min`` and ``max`` and the plan is rejected if any value
    is out of range. Range validation is inteded primarily for using with parameters that
    accept numbers or lists/tuples of numbers. The ``step`` value specified in annotation
    is not used by Queue Server, but passed to the client application and may be useful for
    generating GUI components (e.g. spin boxes).

    Another potential use of the decorator is to provide shorter text descriptions (e.g. for tooltips)
    for the parameters in case long description extracted from the function docstring is too detailed.
    The plan and parameter descriptions defined in the decorator override the descriptions from
    the docstring, hiding them from the client applications.

    The decorator does not change the function and does not overwrite an existing docstring.
    The decorator does not generate function descriptions, instead the parameter dictionary
    passed to the decorator is saved as ``_custom_parameter_annotation_`` attribute of the function
    and may be used later for generation of plan descriptions.

    The decorator verifies if the parameter dictionary matches JSON schema and if names of all
    the parameter names exist in the function signature. The exception is raised if there is
    a mismatch.

    The following is an example of custom annotation (Python dictionary) with inline comments.
    The example is not practical. It demonstrates syntax of defining various annotation elements
    using the decorator:

    .. code-block:: python

        from bluesky_queueserver import parameter_annotation_decorator

        @parameter_annotation_decorator({
            # Optional function description.
            "description": "Custom annotation with plans and devices.",

            # 'parameters' dictionary is optional. Empty 'parameters' dictionary will
            #   also be accepted. The keys in the parameter dictionary are names of
            #   the function parameters. The parameters must exist in function signature,
            #   otherwise an exception will be raised.
            "parameters": {

                "experiment_name": {
                    # This parameter will probably have type hint 'some_name: str`.
                    "description": "String selected from a list of strings (similar to enum)",
                    "annotation": "Names",

                    # Note, that "devices", "plans" and "names" are treated identically
                    #   during type checking, since they represent lists of strings.
                    #   One parameter may have name groups from "devices", "plans"
                    #   and "enums" combined in complex expression using 'typing'
                    #   module. The names listed as ``enums`` are not replaced by
                    #   references to objects in RE Worker namespace.
                    "enums": {
                        "Names": ("name1", "name2", "name3"),
                    },
                }

                "devices": {
                    "description": "Parameter that accepts the list of devices.",
                    "annotation": "typing.List[typing.Union[DeviceType1, DeviceType2]]",

                    # Here we provide the list of devices. 'devices' and 'plans' are
                    #    treated similarly, but should be separated into different sections
                    #    to implement proper parameter validation. The devices may be listed
                    #    explicitly by name or using patterns based on regular expressions.
                    #    The list may contain any combination of names and patterns.
                    #    The patterns are distinguished from names by leading ``:`` at the
                    #    beginning of each pattern. Patterns may contain multiple regular
                    #    expressions separated with ``:``: the first expression is applied
                    #    to the device name, the second expression to subdevices of the
                    #    device, the third - to subdevices of subdevices etc. For example,
                    #    the pattern ``:^stage_:^det:val$ would match the devices with names
                    #    such as ``stage_sim.det2.val``.
                    #
                    #    Adding ``-`` before an expression (after ``:``) excludes the matching
                    #    devices/subdevices from the list (e.g. ``:^stage_:-^det:val$ adds
                    #    ``stage_sim`` and ``stage_sim.det2.val`` to the list, but not
                    #    ``stage_sim.det2``).
                    #
                    #    A regular expression may be applied to the full name (including
                    #    subdevice names) or remaining name of the device. An expression
                    #    could be labelled as 'full-name' by placing ``?`` before it.
                    #    For example ``:?^stage_.*val$`` would pick all devices with names
                    #    starting with ``stage_`` and ending with ``val`` from the complete
                    #    tree of existing devices. The search depth may be restricted by
                    #    adding ``depth`` parameter (e.g. ``:?^stage_.*val$:depth=5``
                    #    restricts the search depth to 5).
                    #
                    #    The both types of expressions could be mixed in one pattern.
                    #    The full-name expression must be the last in the pattern and
                    #    is applied to the remaining part of the name. For example
                    #    ``:^stage_:?^det.*val$:depth=4`` selects the device ``stage_sim``
                    #    and all its subdevices starting with ``det`` and ending with ``val``
                    #    up to the total depth of 5 (depth=4 is set for the full-name expression).
                    #    Note, that searching devices using full-name expressions is less
                    #    efficient and should be minimized or avoided if possible.
                    #
                    #    Regular expressions may be preceded with one the keywords that
                    #    defines the type of matching devices: ``__READABLE__``,
                    #    ``__MOTOR__``, ``__DETECTOR__`` and ``__FLYABLE__``.
                    #    For example, ``__READABLE__:^stage_:?^det.*val$:depth=4``
                    #    selects only the devices that are readable.
                    "devices": {
                        "DeviceType1": ("det1", ":^det2$", ":^det:val$", ":?^det.*val$"),
                        "DeviceType2": ("__MOTOR__:?.*:depth=5"),
                    },
                    # Set default value to string 'det1'. The default value MUST be
                    #   defined in the function header for the parameter that has
                    #   the default value overridden in the decorator. The default
                    #   value in the header is reference to 'det1'.
                    "default": "'det1'",
                },

                # Parameter that accepts a plan or a list of plans
                #   (plans must exist in RE Worker namespace,
                #    reference to the plans will be passed to the plan).
                "plans_to_run": {
                    # Text descriptions of parameters are optional.
                    "description": "Parameter that accepts a plan or a list of plans.",
                    "annotation": "typing.Union[PlanType1, typing.List[PlanType2]]",

                    # Similarly to lists of name patterns for devices, the lists of
                    #   patterns for plan may include device names and name patterns
                    #   based on regular expressions. The patterns always start with
                    #   ``:`` and may contain only a single regular expressions.
                    #   The regular expression may be preced with ``+``, ``-`` and ``?``
                    #   characters (e.g. ``:?^count$), but those characters are ignored
                    #   by the processing algorithm. ``depth`` can not be used in
                    #   the plan patterns.
                    #
                    # In the following example, two lists of plan names are defined.
                    #   Names of the plan lists are used as types in 'annotation'.
                    #   The example of annotation above allows to pass one plan from
                    #   the group 'PlanType1' or a list of plans from 'PlanType2'.
                    "plans": {
                        "PlanType1": (":^count$", "scan", "gridscan"),
                        "PlanType2": ("more", "plan", "names", ":^move_"),
                    },
                },

                "dwell_time": {
                    "description": "Dwell time.",
                    # The simple type 'float' should probably be specified as
                    #   a parameter hint, but it can also be specified here. Putting
                    #   it in both places will also work.
                    "annotation": "float",
                    "min": 0.1,
                    "max": 10.0,
                    "step": 0.1,
                }

                "str_or_int_or_float": {
                    "description": "Some values that may be of 'str', 'int' or 'float' type.",
                    # The following expression can also be put as the parameter hint.
                    "annotation": "typing.Union[str, int, float]",
                }

            },
        })
        def plan(experiment_name, plans_to_run, devices=det1, dwell_time=1.0, str_or_int_or_float=5):
            <code implementing the plan>

    Raises
    ------
    jsonschema.ValidationError
        The parameter dictionary does not match the schema.
    ValueError
        The dictionary contains parameters that are not function arguments (not in the function signature).
    """

    def function_wrap(func):
        if inspect.isgeneratorfunction(func):

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return (yield from func(*args, **kwargs))

        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

        # Always create the copy (annotation dictionary may be reused)
        nonlocal annotation
        annotation = copy.deepcopy(annotation)

        # Create custom validator which recognizes lists and tuples as jsonschema 'arrays'
        def is_array(checker, instance):
            return isinstance(instance, (list, tuple))

        type_checker = jsonschema.Draft3Validator.TYPE_CHECKER.redefine("array", is_array)
        ValidatorCustom = jsonschema.validators.extend(jsonschema.Draft7Validator, type_checker=type_checker)
        validator = ValidatorCustom(schema=_parameter_annotation_schema)

        # Validate annotation
        validator.validate(annotation)

        sig = inspect.signature(func)
        parameters = sig.parameters

        param_unknown = []
        if "parameters" in annotation:
            for p in annotation["parameters"]:
                if p not in parameters:
                    param_unknown.append(p)
        if param_unknown:
            msg = (
                f"Custom annotation parameters {param_unknown} are not "
                f"in the signature of function '{func.__name__}'."
            )
            raise ValueError(msg)

        setattr(wrapper, "_custom_parameter_annotation_", annotation)

        return wrapper

    return function_wrap



import typing
import pytest
import jsonschema
import inspect
from bluesky_queueserver.manager.annotation_decorator import (
    parameter_annotation_decorator,
)


_simple_annotation = {
    "description": "Simple generator function",
    "parameters": {
        "val_arg": {"description": "Parameter 'val_arg'"},
        "val_kwarg": {"description": "Parameter 'val_kwarg'"},
    },
}


_simple_annotation_no_fdesc = {
    "parameters": {
        "val_arg": {"description": "Parameter 'val_arg'"},
        "val_kwarg": {"description": "Parameter 'val_kwarg'"},
    },
}


_simple_annotation_no_pdesc = {
    "description": "Simple generator function",
    "parameters": {
        "val_arg": {},
        "val_kwarg": {},
    },
}


# Note: the types in the following annotation don't match the types in the function
# parameter annotation on purpose. It allows to verify which types are actually placed
# in the docstring.
_simple_annotation_with_types = {
    "description": "Simple function with type annotations",
    "parameters": {
        "val_arg": {"description": "Parameter 'val_arg'", "annotation": "float"},
        "val_kwarg": {"description": "Parameter 'val_kwarg'", "annotation": "int"},
    },
}


_simple_annotation_list_devices_and_plans = {
    "description": "Simple function with type annotations",
    "parameters": {
        "val_arg": {
            "description": "Parameter 'val_arg'",
            "annotation": "Motor",
            "devices": {
                "Motor": ("m1", "m2", "m3"),
            },
        },
        "val_kwarg": {
            "description": "Parameter 'val_kwarg'",
            "annotation": "Plan",
            "plans": {
                "Plan": ("p1", "p2", "p3"),
            },
        },
    },
}


_simple_annotation_enums = {
    "description": "Simple function with type annotations",
    "parameters": {
        "val_arg": {
            "description": "Parameter 'val_arg'",
            "annotation": "typing.List[CustomEnum]",
            "enums": {
                "CustomEnum": ("m1", "m2", "m3"),
            },
        },
        "val_kwarg": {
            "description": "Parameter 'val_kwarg'",
            "annotation": "CustomEnum",
            "enums": {
                "CustomEnum": ("p1", "p2", "p3"),
            },
        },
    },
}


_simple_annotation_with_default = {
    "description": "Simple function with type annotations",
    "parameters": {
        "val_arg": {
            "description": "Parameter 'val_arg'",
            "annotation": "int",
            "default": 10,
        },
        "val_kwarg": {
            "description": "Parameter 'val_kwarg'",
            "annotation": "CustomEnum",
            "enums": {
                "CustomEnum": ("p1", "p2", "p3"),
            },
            "default": "p1",  # String value should be quoted
        },
    },
}

_simple_annotation_with_min_max_step = {
    "description": "Simple function with type annotations",
    "parameters": {
        "val_arg": {
            "description": "Parameter 'val_arg'",
            "default": 10,
            "min": 1,
            "max": 100,
            "step": 0.1,
        },
    },
}


_simple_annotation_long_descriptions = {
    "description": "Simple function with type annotations and long description. " * 10,
    "parameters": {
        "val_arg": {"description": "Long description of the parameter 'val_arg'. " * 10, "annotation": "float"},
        "val_kwarg": {
            "description": "Long description of the parameter 'val_kwarg'. " * 10,
            "annotation": "int",
        },
    },
}


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    {},  # Empty custom annotation is allowed
    _simple_annotation,
    _simple_annotation_no_fdesc,
    _simple_annotation_no_pdesc,
    _simple_annotation_with_types,
    _simple_annotation_list_devices_and_plans,
    _simple_annotation_enums,
    _simple_annotation_with_default,
    _simple_annotation_with_min_max_step,
    _simple_annotation_long_descriptions,
])
# fmt: on
def test_annotation_dectorator_01(custom_annotation):
    """
    Basic test: decorator is applied to a simple function
    """

    @parameter_annotation_decorator(custom_annotation)
    def func(val_arg, *, val_kwarg=10):
        return int(val_arg + val_kwarg)

    assert func._custom_parameter_annotation_ == custom_annotation
    assert func(10, val_kwarg=20) == 30
    assert func.__name__ == "func"


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    _simple_annotation_with_types,
    _simple_annotation_long_descriptions,

])
# fmt: on
def test_annotation_dectorator_02(custom_annotation):
    """
    Types are specified in the custom annotation and in Python parameter annotations.
    Python annotations are not used. Verify that the signature was not broken.
    """

    @parameter_annotation_decorator(custom_annotation)
    def func(val_arg: int, *, val_kwarg: float = 10):
        return int(val_arg + val_kwarg)

    assert func._custom_parameter_annotation_ == custom_annotation
    assert func(10, val_kwarg=20) == 30
    assert func.__name__ == "func"


_simple_annotation_no_types = {
    "description": "Simple function with type annotations",
    "parameters": {
        "val_arg": {"description": "Parameter 'val_arg'"},
        "val_kwarg": {"description": "Parameter 'val_kwarg'"},
    },
}


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    _simple_annotation_no_types,
])
# fmt: on
def test_annotation_dectorator_03(custom_annotation):
    """
    Types are specified in the custom annotation and in Python parameter annotations.
    Python annotations are not used.
    """

    @parameter_annotation_decorator(custom_annotation)
    def func(val_arg: int, *, val_kwarg: float = 10) -> int:
        return int(val_arg + val_kwarg)

    assert func._custom_parameter_annotation_ == custom_annotation
    assert func(10, val_kwarg=20) == 30
    assert func.__name__ == "func"


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    _simple_annotation,
])
# fmt: on
def test_annotation_dectorator_04(custom_annotation):
    """
    Basic test: decorator is applied to a generator function.
    """

    @parameter_annotation_decorator(custom_annotation)
    def func(val_arg, *, val_kwarg=10):
        n_elements = val_arg + val_kwarg
        for n in range(n_elements):
            yield n

    assert func._custom_parameter_annotation_ == custom_annotation
    assert list(func(10, val_kwarg=20)) == list(range(30))
    assert func.__name__ == "func"


_custom_annotation_plans_and_devices = {
    "description": "Custom annotation with plans and devices.",
    "parameters": {
        "val_arg": {
            "description": "Parameter that accepts a single plan.",
            "annotation": "typing.Union(Plan1, Plan2)",
            "plans": {"Plan1": ("count", "scan", "gridscan"), "Plan2": ("some", "other", "plans")},
        },
        "val_kwarg": {
            "description": "Parameter that accepts the list of devices.",
            "annotation": "typing.List(Device)",
            "devices": {"Device": ("det1", "det2", "det3")},
        },
    },
}


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    _custom_annotation_plans_and_devices,
])
# fmt: on
def test_annotation_dectorator_05(custom_annotation):
    """
    Test if the list of plans and devices is displayed correctly in the docstring.
    """

    @parameter_annotation_decorator(custom_annotation)
    def func(val_arg, *, val_kwarg) -> typing.Iterable[int]:
        # The following code has no purpose except to see if the generator is still working.
        added_strings = val_arg + val_kwarg
        for n in range(len(added_strings)):
            yield n

    assert func._custom_parameter_annotation_ == custom_annotation
    assert list(func("gridscan", val_kwarg="det2")) == list(range(len("gridscan") + len("det2")))
    assert func.__name__ == "func"


_more_complicated_annotation = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "detector": {
            "devices": {"Device": ("det1", "det2", "det3")},
            "annotation": "Device",
            "description": "Device name",
        },
        "detectors": {
            "devices": {"Device": ("det1", "det4"), "Motor": ("motor10", "motor12")},
            "annotation": "typing.List[typing.Union[Device, Motor]]",
            "description": "Device names",
        },
        "args": {
            "devices": {"Motor": ("motor1", "motor2", "motor3"), "Detector": ("det30", "det31")},
            "annotation": "typing.Union[Detector, Motor, int]",
            "description": "Motors or ints",
        },
        "kwargs": {
            "devices": {"Detector": ("det50", "det51")},
            "annotation": "typing.Union[float, Detector]",
            "description": "Detectors or floats",
        },
    },
}


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    _more_complicated_annotation,
])
# fmt: on
def test_annotation_dectorator_06(custom_annotation):
    """
    An example of more complicated docstring with varargs and varkwargs. Some of the
    args and kwargs are not documented.
    """

    @parameter_annotation_decorator(custom_annotation)
    def func(
        detector: typing.Any,
        detectors: typing.List[typing.Any],
        val1: float = 10,
        *args: int,
        msg: str = "default_string",
        val2: typing.Union[int, float] = 6,
        **kwargs: int,
    ) -> typing.Generator[str, None, None]:
        str_list = ["str1", "str2", "str3"]
        for s in str_list:
            yield s

    assert func._custom_parameter_annotation_ == custom_annotation
    assert list(func("det1", ["det1", "det2"])) == ["str1", "str2", "str3"]
    assert func.__name__ == "func"


_trivial_annotation = {
    "description": "Example of annotation with varargs and varkwargs.",
}

# _trivial_annotation_doc = """Example of annotation with varargs and varkwargs.
# """


# fmt: off
@pytest.mark.parametrize("custom_annotation", [
    _trivial_annotation,
])
# fmt: on
def test_annotation_dectorator_07(custom_annotation):
    """
    The trivial case of the decorator that only specifies function description
    and the function has no parameters.
    """

    @parameter_annotation_decorator(custom_annotation)
    def func():
        str_list = ["str1", "str2", "str3"]
        for s in str_list:
            yield s

    assert func._custom_parameter_annotation_ == custom_annotation
    assert list(func()) == ["str1", "str2", "str3"]
    assert func.__name__ == "func"


_annotation_with_convert_plan_and_device_names = {
    "description": "Example of using 'convert_strings_to_objects'.",
    "parameters": {
        "existing_param": {
            "description": "The list of strings that should be converted to plans or devices.",
            "annotation": "typing.List[str]",
            "convert_plan_names": True,
            "convert_device_names": True,
        }
    },
}


def test_annotation_dectorator_08():
    """
    The trivial case of the decorator that only specified function description
    and the function has no parameters.
    """

    @parameter_annotation_decorator(_annotation_with_convert_plan_and_device_names)
    def func(existing_param):
        pass

    assert func._custom_parameter_annotation_ == _annotation_with_convert_plan_and_device_names


_trivial_annotation_error1 = {
    "descriptions": "Example of annotation with varargs and varkwargs.",
}

_trivial_annotation_error2 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {"no_such_parameter": {"description": "The function has no such parameter"}},
}

_trivial_annotation_error3 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {"descriptions": "Required key is 'discription'. Schema validation should fail."}
    },
}

_trivial_annotation_error4 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "Motor",
            "some_devices": {  # No such keyword
                "Motor": ("a", "b", "c"),
            },
        }
    },
}

_trivial_annotation_error5 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "Motor",
            "devices": {
                "Motor": (1, 2, 3),  # Type must be 'str'
            },
        }
    },
}

_trivial_annotation_error6 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "Motor",
            "devices": {
                "Motor": "unsupported_list",  # Incorrect name for built-in list
            },
        }
    },
}

_trivial_annotation_error7 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "Motor",
            "devices": {
                "Motor": None,  # Incorrect name for built-in list
            },
        }
    },
}

_trivial_annotation_error8 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "Plans",
            "plans": {
                "Plans.": "someplan",  # Plan name 'Plans.' contains character '.' which is not allowed
            },
        }
    },
}

_trivial_annotation_error9 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "str",
            "convert_plan_names": "abc",  # Must be boolean
        }
    },
}

_trivial_annotation_error10 = {
    "description": "Example of annotation with varargs and varkwargs.",
    "parameters": {
        "existing_param": {
            "description": "Required key is 'discription'. Schema validation should fail.",
            "annotation": "str",
            "convert_device_names": 10,  # Must be boolean
        }
    },
}


# fmt: off
@pytest.mark.parametrize("custom_annotation, ex_type, err_msg", [
    (_trivial_annotation_error1, jsonschema.ValidationError,
     r"Additional properties are not allowed \('descriptions' was unexpected\)"),
    (_trivial_annotation_error2, ValueError, r"\['no_such_parameter'] are not in the signature of function"),
    (_trivial_annotation_error3, jsonschema.ValidationError,
     r"Additional properties are not allowed \('descriptions' was unexpected\)"),
    (_trivial_annotation_error4, jsonschema.ValidationError,
     r"Additional properties are not allowed \('some_devices' was unexpected\)"),
    (_trivial_annotation_error5, jsonschema.ValidationError, r"\(1, 2, 3\) is not valid"),
    (_trivial_annotation_error6, jsonschema.ValidationError, "'unsupported_list' is not valid"),
    (_trivial_annotation_error7, jsonschema.ValidationError, "None is not valid under any of the given schemas"),
    (_trivial_annotation_error8, jsonschema.ValidationError, "'Plans.' does not match any of the regexes"),
    (_trivial_annotation_error9, jsonschema.ValidationError, "'abc' is not of type 'boolean'"),
    (_trivial_annotation_error10, jsonschema.ValidationError, "10 is not of type 'boolean'"),
])
# fmt: on
def test_annotation_dectorator_09_fail(custom_annotation, ex_type, err_msg):
    """
    The trivial case of the decorator that only specified function description
    and the function has no parameters.
    """
    with pytest.raises(ex_type, match=err_msg):

        @parameter_annotation_decorator(custom_annotation)
        def func(existing_param):
            pass


_trivial_annotation = {
    "description": "Trivial annotation.",
}


def test_annotation_dectorator_10():
    """
    Test if decorated generator function is recognized as a generator function by ``inspect``.
    """
    # Apply decorator to a function
    @parameter_annotation_decorator(_trivial_annotation)
    def func():
        return None

    assert inspect.isgeneratorfunction(func) is False

    # Apply decorator to a generator function
    @parameter_annotation_decorator(_trivial_annotation)
    def func():
        yield from [1, 2, 3]

    assert inspect.isgeneratorfunction(func) is True



import argparse
import asyncio
import inspect
import io
import json
import sys
import time as ttime
import threading
import zmq

import bluesky_queueserver

import logging

logger = logging.getLogger(__name__)
qserver_version = bluesky_queueserver.__version__


class ConsoleOutputStream(io.TextIOBase):
    """
    Class that implements writable text file object that collects printed console messages
    and adds timestamps to messages and adds the message to the queue. The messages are
    dictionaries in the form ``{"time": <timestamp>, "msg": <printed text>}.

    Parameters
    ----------
    msg_queue : multiprocessing.Queue
        Reference to the queue used for collecting messages.
    """

    def __init__(self, *, msg_queue):
        super().__init__()
        self._msg_queue = msg_queue
        self._stdout = sys.__stdout__

    def write(self, s):
        """
        Overrides the method of ``io.TextIOBase``.
        """
        s = str(s)
        msg = {"time": ttime.time(), "msg": s}
        self._msg_queue.put(msg)
        return len(s)


def redirect_output_streams(file_obj):
    """
    Override the default output streams with custom file object.
    The object may be an instance of ``ConsoleOutputStream``.

    Parameters
    ----------
    file_obj : ConsoleOutputStream
        Reference for the open writable file object (text output).
    """
    sys.stdout = file_obj
    sys.stderr = file_obj


def setup_console_output_redirection(msg_queue):
    """
    Set up redirection of console output. If ``msg_queue`` is ``None``, then do nothing.

    Parameters
    ----------
    msg_queue : multiprocessing.Queue
        Queue that is used to collect console output messages.
    """
    if msg_queue:
        fobj = ConsoleOutputStream(msg_queue=msg_queue)
        redirect_output_streams(fobj)


_default_zmq_console_topic = "QS_Console"


class PublishConsoleOutput:
    """
    The class that is publishing the collected console output messages to 0MQ socket.
    The queue is expected to be filled with messages in the format
    ``{"time": <timestamp>, "msg": <text message>}``. The object of the class
    receives the reference to the queue during initialization. The collected messages
    are published as they are added to the queue. The messages may be collected
    in multiple processes.

    Parameters
    ----------
    msg_queue : multiprocessing.Queue
        Reference to the queue object, used for collecting of the output messages.
        The messages added to the queue will be automatically published to 0MQ socket.
    console_output_on : boolean
        Enable/disable printing console output to the terminal
    zmq_publish_on : boolean
        Enable/disable publishing console output to 0MQ socket
    zmq_publish_addr : str, None
        Address of 0MQ PUB socket for the publishing server. If ``None``, then
        the default address ``tcp://*:60625`` is used.
    zmq_topic : str
        Name of the 0MQ topic where the messages are published.
    name : str
        Name of the thread where the messages are published.
    """

    def __init__(
        self,
        *,
        msg_queue,
        console_output_on=True,
        zmq_publish_on=True,
        zmq_publish_addr=None,
        zmq_topic=_default_zmq_console_topic,
        name="RE Console Output Publisher",
    ):
        self._thread_running = False  # Set True to exit the thread
        self._thread_name = name
        self._msg_queue = msg_queue
        self._polling_timeout = 0.1  # in sec.

        self._console_output_on = console_output_on
        self._zmq_publish_on = zmq_publish_on

        zmq_publish_addr = zmq_publish_addr or "tcp://*:60625"

        self._zmq_publish_addr = zmq_publish_addr
        self._zmq_topic = zmq_topic

        self._socket = None
        if self._zmq_publish_on:
            try:
                context = zmq.Context()
                self._socket = context.socket(zmq.PUB)
                self._socket.bind(self._zmq_publish_addr)
            except Exception as ex:
                logger.error(
                    "Failed to create 0MQ socket at %s. Console output will not be published. Exception: %s",
                    self._zmq_publish_addr,
                    str(ex),
                )

        if self._socket and self._zmq_publish_on:
            logging.info("Publishing console output to 0MQ socket at %s", zmq_publish_addr)

    def start(self):
        """
        Start thread polling the queue.
        """
        self._start_processing_thread()

    def stop(self):
        """
        Stop thread that polls the queue (and exit the tread)
        """
        self._thread_running = False

    def __del__(self):
        self.stop()
        if self._socket:
            self._socket.close()

    def _start_processing_thread(self):
        # The thread should not be started of Message Queue object does not exist
        if not self._thread_running and self._msg_queue:
            self._thread_running = True
            self._thread_conn = threading.Thread(
                target=self._publishing_thread, name=self._thread_name, daemon=True
            )
            self._thread_conn.start()

    def _publishing_thread(self):
        while True:
            try:
                msg = self._msg_queue.get(block=True, timeout=self._polling_timeout)
                self._publish(msg)
            except Exception:
                pass

            if not self._thread_running:  # Exit thread
                break

    def _publish(self, payload):
        if self._console_output_on:
            sys.__stdout__.write(payload["msg"])
            sys.__stdout__.flush()

        if self._zmq_publish_on and self._socket:
            topic = self._zmq_topic
            payload_json = json.dumps(payload)
            self._socket.send_multipart([topic.encode("ascii"), payload_json.encode("utf8")])


class ReceiveConsoleOutput:
    """
    The class allows to subscribe to published 0MQ messages and read the messages one by
    one as they arrive. Subscription is performed using the remote 0MQ address and topic.

    The class provides blocking (with timeout) ``recv`` method that waits for the next
    published message. The following example contains the code illustrating using the class.
    In real-world  application the loop will be running in a separate thread and generating
    callbacks on each received message.

    .. code-block:: python

        from bluesky_queueserver import ReceiveConsoleOutput

        zmq_subscribe_addr = "tcp://localhost:60625"
        rco = ReceiveConsoleOutput(zmq_subscribe_addr=zmq_subscribe_addr)

        while True:
            try:
                payload = rco.recv()
                time, msg = payload.get("time", None), payload.get("msg", None)
                # In this example the messages are printed in the terminal.
                sys.stdout.write(msg)
                sys.stdout.flush()
            except TimeoutError:
                # Timeout does not mean communication error!!!
                # Insert the code that needs to be executed on timeout (if any).
                pass
            # Place for the code that should be executed after receiving each
            #   message or after timeout (e.g. check a condition and exit
            #   the loop once the condition is satisfied).


    Parameters
    ----------
    zmq_subscribe_addr : str or None
        Address of ZMQ server (PUB socket). If None, then the default address is
        ``tcp://localhost:60625`` is used.
    zmq_topic : str
        0MQ topic for console output. Only messages from this topic are going to be received.
    timeout : int, float or None
        Timeout for the receive operation in milliseconds. If `None`, then wait
        for the message indefinitely.
    """

    def __init__(self, *, zmq_subscribe_addr=None, zmq_topic=_default_zmq_console_topic, timeout=1000):

        self._timeout = timeout  # Timeout for 'recv' operation (ms)

        zmq_subscribe_addr = zmq_subscribe_addr or "tcp://localhost:60625"

        logger.info(f"Subscribing to console output stream from 0MQ address: {zmq_subscribe_addr} ...")
        logger.info(f"Subscribing to 0MQ topic: '{zmq_topic}' ...")
        self._zmq_subscribe_addr = zmq_subscribe_addr
        self._zmq_topic = zmq_topic

        self._socket = None
        if self._zmq_subscribe_addr:
            context = zmq.Context()
            self._socket = context.socket(zmq.SUB)
            self._socket.connect(self._zmq_subscribe_addr)
            self._socket.subscribe(self._zmq_topic)

    def recv(self, timeout=-1):
        """
        Get the next published message. If timeout expires then
        ``TimeoutError`` is raised.

        Parameters
        ----------
        timeout : int, float or None
            Timeout for the receive operation in milliseconds. If timeout is
            a negative number (default), the timeout value passed to the class
            constructor is used. If `None`, then wait indefinitely.

        Returns
        -------
        dict
            Received message. The dictionary contains timestamp (``time`` key)
            and text message (``msg`` key).

        Raises
        ------
        TimeoutError
            Timeout occurred. Timeout does not indicate communication error.
        """

        if (timeout is not None) and (timeout < 0):
            timeout = self._timeout

        if not self._socket.poll(timeout=timeout):
            raise TimeoutError("No message received during timeout period {timeout} ms")

        topic, payload_json = self._socket.recv_multipart()
        payload_json = payload_json.decode("utf8", "strict")
        payload = json.loads(payload_json)
        return payload

    def __del__(self):
        self._socket.close()


class ReceiveConsoleOutputAsync:
    """
    Async version of ``ReceiveConsoleOutput`` class. There are two ways to use the class:
    explicitly awaiting for the ``recv`` function (same as in ``ReceiveConsoleOutput``)
    or setting up a callback function (plain function or coroutine).

    Explicitly awaiting ``recv`` function:

    .. code-block:: python

        from bluesky_queueserver import ReceiveConsoleOutputAsync

        zmq_subscribe_addr = "tcp://localhost:60625"
        rco = ReceiveConsoleOutputAsync(zmq_subscribe_addr=zmq_subscribe_addr)

        async def run_acquisition():
            while True:
                try:
                    payload = await rco.recv()
                    time, msg = payload.get("time", None), payload.get("msg", None)
                    # In this example the messages are printed in the terminal.
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                except TimeoutError:
                    # Timeout does not mean communication error!!!
                    # Insert the code that needs to be executed on timeout (if any).
                    pass
                # Place for the code that should be executed after receiving each
                #   message or after timeout (e.g. check a condition and exit
                #   the loop once the condition is satisfied).

        asyncio.run(run_acquisition())

    Setting up callback function or coroutine (awaitable function):

    .. code-block:: python

        from bluesky_queueserver import ReceiveConsoleOutputAsync

        zmq_subscribe_addr = "tcp://localhost:60625"
        rco = ReceiveConsoleOutputAsync(zmq_subscribe_addr=zmq_subscribe_addr)

        async def cb_coro(payload):
            time, msg = payload.get("time", None), payload.get("msg", None)
            # In this example the messages are printed in the terminal.
            sys.stdout.write(msg)
            sys.stdout.flush()

        rco.set_callback(cb_coro)

        async def run_acquisition():
            rco.start()
            # Do something useful here, e.g. sleep
            asyncio.sleep(60)
            rco.stop()

            # Acquisition can be started and stopped multiple time if necessary
            rco.start()
            asyncio.sleep(60)
            rco.stop()

        asyncio.run(run_acquisition())

    .. note::
        If callback is a plain function, it is executed immediately after the message is received
        and may potentially block the loop if it takes too long to complete (even occasionally).
        If the callback is a coroutine, it is not awaited, but instead placed in the loop
        (with ``ensure_future``), so acquisition of messages will continue. Typically the callback
        will do a simple operation such as adding the received message to the queue.

    Parameters
    ----------
    zmq_subscribe_addr : str or None
        Address of ZMQ server (PUB socket). If None, then the default address is
        ``tcp://localhost:60625`` is used.
    zmq_topic : str
        0MQ topic for console output. Only messages from this topic are going to be received.
    timeout : int, float or None
        Timeout for the receive operation in milliseconds. If `None`, then wait
        for the message indefinitely.
    """

    def __init__(self, *, zmq_subscribe_addr=None, zmq_topic=_default_zmq_console_topic, timeout=1000):

        self._timeout = timeout  # Timeout for 'recv' operation (ms)

        zmq_subscribe_addr = zmq_subscribe_addr or "tcp://localhost:60625"

        self._callback = None  # Function that is awaited once a message is received from RE Manager
        self._exit = False
        self._is_running = False

        logger.info(f"Subscribing to console output stream from 0MQ address: {zmq_subscribe_addr} ...")
        logger.info(f"Subscribing to 0MQ topic: '{zmq_topic}' ...")
        self._zmq_subscribe_addr = zmq_subscribe_addr
        self._zmq_topic = zmq_topic

        self._socket = None
        if self._zmq_subscribe_addr:
            context = zmq.asyncio.Context()
            self._socket = context.socket(zmq.SUB)
            self._socket.connect(self._zmq_subscribe_addr)

    def set_callback(self, cb):
        """
        Set callback function, which is called once for each received message. If ``cb`` is
        a function, it is called immediately and execution of the loop is blocked until the
        execution of the function is complete. If ``cb`` is coroutine, it is not awaited, but
        instead placed in the loop using ``asyncio.ensure_future``. Only one callback function
        can be set.

        Parameters
        ----------
        cb : callable, coroutine or None
            Reference to a callback function or coroutine. The function signature is expected
            to receive a message as a parameter (message is a dictionary with keys ``time`` and ``msg``)
            and return ``None``. The function is expected to handle exceptions that are raised
            internally. Pass ``None`` to clear callback (messages will be received and discarded).
        """
        self._callback = cb

    async def recv(self, timeout=-1):
        """
        Get the next published message. If timeout expires then ``TimeoutError`` is raised.

        Parameters
        ----------
        timeout : int, float or None
            Timeout for the receive operation in milliseconds. If timeout is
            a negative number (default), the timeout value passed to the class
            constructor is used. If `None`, then wait indefinitely.

        Returns
        -------
        dict
            Received message. The dictionary contains timestamp (``time`` key)
            and text message (``msg`` key).

        Raises
        ------
        TimeoutError
            Timeout occurred. Timeout does not indicate communication error.
        """

        if (timeout is not None) and (timeout < 0):
            timeout = self._timeout

        if not await self._socket.poll(timeout=timeout):
            raise TimeoutError("No message received during timeout period {timeout} ms")

        topic, payload_json = await self._socket.recv_multipart()

        payload_json = payload_json.decode("utf8", "strict")
        payload = json.loads(payload_json)
        return payload

    async def _recv_next_message(self):
        try:
            payload = await self.recv()
            if self._callback:
                if inspect.iscoroutinefunction(self._callback):
                    asyncio.ensure_future(self._callback(payload))
                else:
                    self._callback(payload)
        except TimeoutError:
            pass
        except Exception as ex:
            logger.exception(f"Exception occurred while while waiting for RE Manager console output message: {ex}")

        if not self._exit:
            asyncio.ensure_future(self._recv_next_message())
        else:
            self._socket.unsubscribe(self._zmq_topic)
            self._is_running = False

    def start(self):
        """
        Start collection of messages published by RE Manager. Collection may be started and stopped
        multiple times during a session. Repeated calls to the ``start`` method are ignored.
        The function MUST be called from the event loop.
        """
        self._exit = False
        if not self._is_running:
            self._is_running = True
            self._socket.subscribe(self._zmq_topic)
            asyncio.ensure_future(self._recv_next_message())

    def stop(self):
        """
        Stop collection of messages published by RE Manager. Call to ``stop`` method unsubscribes
        the client from 0MQ topic, therefore all the messages published until collection is started
        are ignored. The function MUST be called from the event loop.
        """
        self._exit = True

    def __del__(self):
        self.stop()
        if self._socket:
            self._socket.close()


def qserver_console_monitor_cli():
    """
    CLI tool for remote monitoring of console output from RE Manager. The function is also
    expected to be used as an example of using  ``ReceiveConsoleOutput`` class.
    """

    logging.basicConfig(level=logging.WARNING)
    logging.getLogger("bluesky_queueserver").setLevel("INFO")

    def formatter(prog):
        # Set maximum width such that printed help mostly fits in the RTD theme code block (documentation).
        return argparse.RawDescriptionHelpFormatter(prog, max_help_position=20, width=90)

    parser = argparse.ArgumentParser(
        description="Queue Server Console Monitor:\nCLI tool for remote monitoring of console output "
        f"published by RE Manager.\nbluesky-queueserver version {qserver_version}\n",
        formatter_class=formatter,
    )
    parser.add_argument(
        "--zmq-subscribe-addr",
        dest="zmq_subscribe_addr",
        type=str,
        default="tcp://localhost:60625",
        help="The address of ZMQ server to subscribe, e.g. 'tcp://127.0.0.1:60625' (default: %(default)s).",
    )

    args = parser.parse_args()
    zmq_subscribe_addr = args.zmq_subscribe_addr

    try:
        rco = ReceiveConsoleOutput(zmq_subscribe_addr=zmq_subscribe_addr)
        while True:
            try:
                payload = rco.recv()
                time, msg = payload.get("time", None), payload.get("msg", None)  # noqa: F841
                sys.stdout.write(msg)
                sys.stdout.flush()
            except TimeoutError:
                # Timeout does not mean communication error!!!
                # There is no need to use or process timeouts. This code
                #   serves mostly as an example of how to use it.
                pass
            # Place for the code that should be executed after receiving each
            #   message or after timeout. (E.g. the code may check some condition
            #   and exit the loop once the condition is fulfilled.)
        exit_code = 0  # The code is set if the loope is exited (which does not happen here)
    except BaseException as ex:
        logger.exception("Queue Server Console Monitor failed with exception: %s", str(ex))
        exit_code = 1
    return exit_code
